AT_AF = "at_af"

base = {
    (0, "at_af"): "núll",
    (1, "et_kk_nf"): "einn",
    (1, "et_kk_þf"): "einn",
    (1, "et_kk_þgf"): "einum",
    (1, "et_kk_ef"): "eins",
    (1, "et_kvk_nf"): "ein",
    (1, "et_kvk_þf"): "eina",
    (1, "et_kvk_þgf"): "einni",
    (1, "et_kvk_ef"): "einnar",
    (1, "et_hk_nf"): "eitt",
    (1, "et_hk_þf"): "eitt",
    (1, "et_hk_þgf"): "einu",
    (1, "et_hk_ef"): "eins",
    (1, "ft_kk_nf"): "einir",
    (1, "ft_kk_þf"): "eina",
    (1, "ft_kk_þgf"): "einum",
    (1, "ft_kk_ef"): "einna",
    (1, "ft_kvk_nf"): "einar",
    (1, "ft_kvk_þf"): "einar",
    (1, "ft_kvk_þgf"): "einum",
    (1, "ft_kvk_ef"): "einna",
    (1, "ft_hk_nf"): "ein",
    (1, "ft_hk_þf"): "ein",
    (1, "ft_hk_þgf"): "einum",
    (1, "ft_hk_ef"): "einna",
    (2, "ft_kk_nf"): "tveir",
    (2, "ft_kk_þf"): "tvo",
    (2, "ft_kk_þgf"): "tveimur/tveim",
    (2, "ft_kk_ef"): "tvegga",
    (2, "ft_kvk_nf"): "tvær",
    (2, "ft_kvk_þf"): "tvær",
    (2, "ft_kvk_þgf"): "tveimur/tveim",
    (2, "ft_kvk_ef"): "tveggja",
    (2, "ft_hk_nf"): "tvö",
    (2, "ft_hk_þf"): "tvö",
    (2, "ft_hk_þgf"): "tveimur/tveim",
    (2, "ft_hk_ef"): "tveggja",
    (3, "ft_kk_nf"): "þrír",
    (3, "ft_kk_þf"): "þrjá",
    (3, "ft_kk_þgf"): "þremur/þrem",
    (3, "ft_kk_ef"): "þriggja",
    (3, "ft_kvk_nf"): "þrjár",
    (3, "ft_kvk_þf"): "þrjár",
    (3, "ft_kvk_þgf"): "þremur/þrem",
    (3, "ft_kvk_ef"): "þriggja",
    (3, "ft_hk_nf"): "þrjú",
    (3, "ft_hk_þf"): "þrjú",
    (3, "ft_hk_þgf"): "þremur/þrem",
    (3, "ft_hk_ef"): "þriggja",
    (4, "ft_kk_nf"): "fjórir",
    (4, "ft_kk_þf"): "fjóra",
    (4, "ft_kk_þgf"): "fjórum",
    (4, "ft_kk_ef"): "fjögurra/fjögra",
    (4, "ft_kvk_nf"): "fjórar",
    (4, "ft_kvk_þf"): "fjórar",
    (4, "ft_kvk_þgf"): "fjórum",
    (4, "ft_kvk_ef"): "fjögurra/fjögra",
    (4, "ft_hk_nf"): "fjögur",
    (4, "ft_hk_þf"): "fjögur",
    (4, "ft_hk_þgf"): "fjórum",
    (4, "ft_hk_ef"): "fjögurra/fjögra",
    (5, "at_af"): "fimm",
    (6, "at_af"): "sex",
    (7, "at_af"): "sjö",
    (8, "at_af"): "átta",
    (9, "at_af"): "níu",
    (10, "at_af"): "tíu",
    (11, "at_af"): "ellefu",
    (12, "at_af"): "tólf",
    (13, "at_af"): "þrettán",
    (14, "at_af"): "fjórtán",
    (15, "at_af"): "fimmtán",
    (16, "at_af"): "sextán",
    (17, "at_af"): "sautján",
    (18, "at_af"): "átján",
    (19, "at_af"): "nítján",
    (20, "at_af"): "tuttugu",
    (30, "at_af"): "þrjátíu",
    (40, "at_af"): "fjörutíu",
    (50, "at_af"): "fimmtíu",
    (60, "at_af"): "sextíu",
    (70, "at_af"): "sjötíu",
    (80, "at_af"): "áttatíu",
    (90, "at_af"): "níutíu",
    (100, "at_af"): "hundrað",
    (1000, "at_af"): "þúsund",
    (1000000, "at_af"): "milljón",
}


declensions = {
    "et": "number",
    "ft": "number",
    "kk": "gender",
    "kvk": "gender",
    "hk": "gender",
    "nf": "case",
    "þf": "case",
    "þgf": "case",
    "ef": "case",
}


def logger():
    import logging

    FORMAT = "[%(filename)s:%(lineno)s - %(funcName)20s()] %(message)s"
    logging.basicConfig(level=logging.WARNING, format=FORMAT)
    logger = logging.getLogger(__name__)
    return logger



from numi.main import spell_out


# Keeping this for quick testing
# for line in spell_out(92):
#     print(line)


# print(spell_out(92))
# for line in spell_out(92, "ft_kvk"):
#     print(line)
# for line in spell_out(92, "kvk"):
#     print(line)
# for line in spell_out(100, "at_af"):
#     print(line)
# for line in spell_out(79, "at_af"):
#     print(line)
# for line in spell_out(13, "at_af"):
#     print(line)
# for line in spell_out(121):
#     print(line)


# for line in spell_out(11):
#     print(line)
# for line in spell_out(111):
#     print(line)
# for line in spell_out(1111):
#     print(line)
# for line in spell_out(9):
#     print(line)
# for line in spell_out(99):
#     print(line)
# for line in spell_out(999):
#     print(line)
# for line in spell_out(9999):
#     print(line)


# print(spell_out(5001, "et_kk_nf"))
# print(spell_out(3124, "ft_kk_nf"))
# print(spell_out(1000, "at_af"))
# print(spell_out(10000))
# print(spell_out(100000))
# print(spell_out(15123))
# print(spell_out(11451))
# print(spell_out(21123))

# 7 print(spell_out(151123))

# print(spell_out(100000))
# print(spell_out(512321))
# print(spell_out(10021, "et_kvk_ef"))
# print(spell_out(10011))
# print(spell_out(10001, "et_kvk_ef"))
# print(spell_out(10221, "et_kvk_ef"))
# print(spell_out(10020))
# print(spell_out(10021, "et_kvk_ef"))
# print(spell_out(10011))
# print(spell_out(10001, "et_kvk_ef"))
# print(spell_out(10221, "et_kvk_ef"))
# for line in spell_out(111101, "et_hk_þf"):
#     print(line)

# print(spell_out(111021, "et_hk_þf"))
# print(spell_out(111001, "et_hk_þf"))
# print(spell_out(111111))
# [
# for line in spell_out(3124, "ft_kk_nf"):
# print(line)
# for line in spell_out(221123, "ft_hk_þgf"):
# print(line)


# print(spell_out("1,1"))
# print(spell_out("2,11"))
# print(spell_out("1,11", "et_kvk_þgf"))
# print(spell_out("1,111", "et_kvk_þgf"))
# print(spell_out("1,1111", "et_kvk_þgf"))


# print(spell_out(975))
# print(spell_out(976))
# print(spell_out(977))
# print(spell_out(978))
# print(spell_out(979))
# print(spell_out(980))
# print(spell_out(92, "ft_kvk_þgf"))
# print(spell_out(995))
# for n in range(1, 1001):
#     for line in spell_out(n):
#         print(line)

# print(spell_out(510000))  # fimmhundruð og tíu þúsund
# print(spell_out(501000))  # fimmhundurð og eitt þúsund
# print(spell_out(503000))  # fimmhundurð og þrjú þúsund
# print(spell_out(212000))  # Tvö hundurð og tólf þúsund
# print(spell_out(914000))  # níuhundurð og fjórtán þúsund
# print(spell_out(18000))  # átján þúsund
# print(spell_out(825000))  # áttahundurð tuttugu og fimm þúsnd



import numbers
from sys import prefix
from tkinter import BOTH
from numi.utils import base, AT_AF, logger
from numi.handle_input import find_valid_combinations, parse_input_string

logging = logger()


def _get_from_base(n, f):
    """
    Returns all possible variations for the written numbers in the base as a list of strings
    for numbers between 1-19 as well as all numbers that are divisable by 10,100,1000 etc.
    """
    num = base[n, f]
    if "/" in num:
        num = num.split("/")
    else:
        num = [num]

    logging.info(f"Returning: {num} Input was: {n, f}")
    return num


def _nums_20_99(n, f):
    """
    Returns all possible variations for the written numbers from 20 to 99 as a list of strings.
    """
    n1 = int(str(n)[0] + "0")
    n2 = int(str(n)[1])
    last_number = _get_from_base(n2, f)

    if len(last_number) == 2:
        num = [f"{base[n1,AT_AF]} og {n2}" for n2 in last_number]
    else:
        num = [f"{base[n1,AT_AF]} og {base[n2,f]}"]

    logging.info(f"Returning: {num} Input was: {n, f}")
    return num


def _nums_100_999(n, f, prefix="both"):
    """
    Returns all possible variations for the written numbers from 100 to 999 as a list of strings.

    prefix values:
        'eitt': Add the word "eitt" infront of "hundrað".
        'hundrað': Return with only "hundrað".
        'both': Return with both "hundrað" and "eitt hundrað".

    """

    def _hundreds(n):
        n_1 = int(str(n)[0])
        if n_1 == 1:
            num = f"{base[n_1, 'et_hk_nf']} hundrað"
        elif n_1 in [2, 3, 4]:
            num = f"{base[n_1,'ft_hk_nf']} hundruð"
        else:
            num = f"{base[n_1,AT_AF]} hundruð"
        logging.info(f"Returning: {num} Input was: {n}")
        return num

    n1 = int(str(n)[0] + "00")
    num1 = _hundreds(n1)
    n2 = n - n1
    num2 = []

    if n in [100, 200, 300, 400, 500, 600, 700, 800, 900]:
        num = [_hundreds(n)]

    elif n2 in [x for x in range(1, 20)] or n2 in [x for x in range(20, 101, 10)]:
        for line in _get_from_base(n2, f):
            num2.append(f"og {line}")
        num2 = num2
    else:
        num2 = _nums_20_99(n2, f)

    if num2:
        num = [f"{num1} {n2}" for n2 in num2]

    if prefix == "both":
        for line in num:
            if "eitt hundrað" in line:
                num.append(line.replace("eitt hundrað", "hundrað"))
    if prefix == "hundrað":
        for idx, line in enumerate(num):
            if "eitt hundrað" in line:
                num[idx] = line.replace("eitt hundrað", "hundrað")
    logging.info(f"Returning: {num} Input was: {n, f}")
    return num


def _nums_1000_9999(n, f, add_thousand=True):
    """
    Returns all possible variations for the written numbers from 1000 to 9999 as a list of strings.
    """

    def _thousands(n):
        n_1 = int(str(n)[0])
        if n_1 == 1:
            num = f"{base[n_1, 'et_hk_nf']} þúsund"
        elif n_1 in [2, 3, 4]:
            num = f"{base[n_1,'ft_hk_nf']} þúsund"
        else:
            num = f"{base[n_1,AT_AF]} þúsund"
        logging.info(f"Returning: {num} Input was: {n}")
        return num

    n1 = int(str(n)[0] + "000")
    num1 = _thousands(n1)
    n2 = n - n1
    num2 = []
    if n in [1000, 2000, 3000, 4000, 5000, 6000, 7000, 8000, 9000]:
        num = [_thousands(n)]
    elif n2 < 1000 and n2 > 100:
        num2 = _nums_100_999(n2, f, prefix="eitt")
    elif n2 in [x for x in range(1, 20)] or n2 in [x for x in range(20, 101, 10)]:
        for line in _get_from_base(n2, f):
            num2.append(f"og {line}")
    else:
        num2 = _nums_20_99(n2, f)

    if num2:
        num = [f"{num1} {n2}" for n2 in num2]

    if n == 1000 and add_thousand:
        for line in num:
            if "eitt þúsund" in line:
                num.append(line.replace("eitt þúsund", "þúsund"))

    logging.info(f"Returning: {num} Input was: {n, f}")
    return num


def _nums_10000_99999(n, f):
    """
    Returns all possible variations for the written numbers from 10000 to 99999 as a list of strings.
    """
    num = []
    num1 = int(str(n)[0:-3])
    num2 = int(str(n)[-3:])
    logging.debug(f"num1: {num1}, num2: {num2}")

    if n in [10000]:
        return [f"{x} þúsund" for x in divert(num1, f)]

    # Numbers 10000-99999
    if num1 in [x for x in range(10, 21)] or num1 in [x for x in range(20, 101, 10)]:
        logging.debug(f"1: {num1}, setting f1 to {AT_AF}")
        f1 = AT_AF
    elif int(str(num1)[-1]) == 1:
        logging.debug(f"2: {num1}, setting f1 to et_hk_nf")
        f1 = "et_hk_nf"

    elif int(str(num1)[-1]) in [2, 3, 4]:
        logging.debug(f"3: {num1}, setting f1 to ft_hk_nf")
        f1 = "ft_hk_nf"
    else:
        logging.debug(f"4: {num1}, setting f1 to {AT_AF}")
        f1 = AT_AF

    if num2 != 0:
        for line in divert(num2, f, prefix="eitt"):
            # Some numbers have an "og" between num1 and num2
            last_two = int(str(num2)[-2:])
            if last_two in [x for x in range(0, 20)] or last_two in [
                x for x in range(20, 91, 10)
            ]:
                num.append(f"{divert(num1, f1, prefix='hundrað')[0]} þúsund og {line}")
            else:
                num.append(f"{divert(num1, f1, prefix='hundrað')[0]} þúsund {line}")
    else:
        logging.debug("num2 is 0")
        num.append(f"{divert(num1, f1, prefix='hundrað')[0]} þúsund")

    logging.info(f"Returning: {num} Input was: {n, f}")
    return num


def _nums_100000_999999(n, f):
    """
    Returns all possible variations for the written numbers from 100000 to 999999 as a list of strings.
    """
    num = []
    num1 = int(str(n)[0:-3])
    num2 = int(str(n)[-3:])

    logging.debug(f"num1: {num1}, num2: {num2}")
    if n in [100000]:
        return [f"{x} þúsund" for x in divert(num1, f)]

    # Numbers 100000-999999
    if int(str(num1)[-2:]) in [x for x in range(10, 20)] or int(str(num1)[-2:]) in [
        x for x in range(20, 101, 10)
    ]:
        logging.debug(f"1: {num1}, setting f1 to {AT_AF}")
        f1 = AT_AF
    elif int(str(num1)[-1]) == 1:
        logging.debug(f"2: {num1}, setting f1 to et_hk_nf")
        f1 = "et_hk_nf"

    elif int(str(num1)[-1]) in [2, 3, 4]:
        logging.debug(f"3: {num1}, setting f1 to ft_hk_nf")
        f1 = "ft_hk_nf"
    else:
        logging.debug(f"4: {num1}, setting f1 to {AT_AF}")
        f1 = AT_AF

    if num2 != 0:
        for n2 in divert(num2, f, prefix="eitt"):
            # Some numbers have an "og" between num1 and num2
            # Numbers 1-9
            if len(str(num2)) == 1:
                logging.info(f'1: Adding "og" {{num2}}')
                num.append(f"{divert(num1, f1, prefix='hundrað')[0]} þúsund og {n2}")
            elif (
                len(str(num2)) == 2
                and num2 in [x for x in range(20, 91, 10)]
                or num2 in range(10, 20)
            ):
                logging.info(f'2: Adding "og" {num2}')
                num.append(f"{divert(num1, f1, prefix='hundrað')[0]} þúsund og {n2}")
            else:
                logging.info(f'3: Not adding "og" {num2}')
                num.append(f"{divert(num1, f1, prefix='hundrað')[0]} þúsund {n2}")
    else:
        logging.debug("num2 is 0")
        num.append(f"{divert(num1, f1, prefix='hundrað')[0]} þúsund")

    logging.info(f"Returning: {num} Input was: {n, f}")
    return num


def add_decimals(nums, nd, f):
    """
    Let's add the decimals to the numbers.
    """
    if len(nums) == 1:
        nums = [[nd, f[0], nums]]

    len_nd = len(str(nd))
    idx = 2
    if len_nd == 1:
        idx = 1

    for line in nums:
        n, f = line[0], line[1]

        f1 = 0
        # if int(str(n)[-idx:]) in [x for x in range(10, 20)] or int(str(n)[-idx:]) in [
        #     x for x in range(20, 101, 10)
        # ]:
        #     f1 = AT_AF
        #     logging.info(f"1: Setting f1 as {f1}")

        # elif int(str(n)[-1]) in [2, 3, 4]:
        #     f1 = "ft_hk_nf"
        #     logging.info(f"2: Setting f1 as {f1}")
        # else:
        #     f1 = AT_AF
        #     logging.info(f"2: Setting f1 as {f1}")
        print("return", n, f1, f": {line[2][0]} komma {divert(n, f)}")
    return nums


def divert(n, f, prefix="both"):
    """
    Handles user input and returns a list of string with all
    possible variations of if input number
    """

    logging.info(f"User input: {n,f}")
    # Numbers 1-19 add 20, 30, 40, 50, 60, 70, 80, 90
    if n in [x for x in range(1, 20)] or n in [x for x in range(20, 91, 10)]:
        return _get_from_base(n, f)
    # Numbers 20 to 99
    elif n in [x for x in range(20, 100)]:
        return _nums_20_99(n, f)
    # Numbers from 100 to 999
    elif len(str(n)) == 3:
        return _nums_100_999(n, f, prefix)
    # Numbers between 1000 and 9999
    elif len(str(n)) == 4:
        return _nums_1000_9999(n, f)
    # Nmnbers between 10000 and 999999
    elif len(str(n)) == 5:
        return _nums_10000_99999(n, f)
    # Nmnbers between 100000 and 9999999
    elif len(str(n)) == 6:
        return _nums_100000_999999(n, f)


def spell_out(n, f=None):
    """
    Validates user input and sends the input to diverter.
    The output is printed to the console.
    """
    logging.info(f"User input: {n,f}")
    f = parse_input_string(f)
    logging.info(
        f"The the len of f after parse input is {len(f)} and of type {type(f)}"
    )
    nd, numbers = None, None
    if "," in str(n):
        logging.info(f"Input contains decimals")
        n, nd = str(n).split(",")
        n, nd = int(n), int(nd)

    f = find_valid_combinations(n, f)

    if len(f) > 1:
        numbers = [[n, v, divert(n, v)] for v in f]
    else:
        numbers = [[n, f[0], divert(n, f[0])]]

    return numbers
    # if nd:
    #     return add_decimals(numbers, nd, f)
    # else:
    #     return numbers



from numi.utils import base, AT_AF, logger
from numi.utils import declensions as dec


logging = logger()


def find_valid_combinations(n, f):
    """
    Given that the last number in the input number controls the declentions
    we check in the base wheter a combination is valid. We return all valid
    combinations. If there are no valid combinations an exception is raised.
    """
    new_f = []

    # if f is only one string at this point, that means the user added a
    # three value string. Let's check if is a valid input.
    # if len(f) == 1:
    #     if (n,f[0]) in base:
    #         logging.debug(f"0: Adding {n,f[0]}")
    #         new_f.append(f[0])
    # if len(str(n)) > 1:
    for v in f:
        # if the whole number along with the input stris in the
        # base, add the combination.
        if (n, v) in base:
            logging.debug(f"1: Adding {n,v}")
            new_f.append(v)
            if n > 1:
                break
        # if the last two digit form a number with the input str
        # that is the base we add that to the combination.
        elif (int(str(n)[-2:]), v) in base:
            logging.debug(f"2: Adding {n,v} based on {int(str(n)[-2:])}")
            new_f.append(v)
            last_two = int(str(n)[-2:])

            if last_two in [x for x in range(10, 20)] or last_two in [
                x for x in range(20, 91, 10)
            ]:
                break

        # if the last number is in the base with the input str
        # that is the base we add that to the combination.
        elif (int(str(n)[-1]), v) in base:
            logging.debug(f"3: Adding {n,v} based on {int(str(n)[-1])}")
            new_f.append(v)

    if new_f:
        return new_f
    else:
        raise Exception(f"Sorry, can't find a valid combination with that input")


def parse_input_string(f):
    """
    This functions takes in the input string and does a few things to make the
    package more robust and user friendly.

    Firstly, it orders the input values separeted by "_" in the correct way. This
    means that the package is agnostic to the order the user sends in.

    Second, if the user does not specify one or more or all of the three input values
    (number, gender, case) we handle that as if the user want all of the possible
    numbers, gender or cases for that input number (number here refers the lingustic number
    i.e. singular or plural but also the input number from the user.)

    Thridly, the function throws an error if the input string has e.g. two or more values
    for one of the three possible (number, gender, case).
    """

    # Create a sorted list of all legal combination found in the base. It's sorted
    # for the sake of testing a set is inherently unsorted
    all_combinations = sorted(list(set([x[1] for x in base])))
    correct_order = ["number", "gender", "case"]

    # If the input string is none we return all possible combinations of
    # the input string as a list of strings
    if f == None:
        logging.debug(f"Returning all combinations as f is {f}")
        return all_combinations

    # If the input string is one of the input numbers that don't follow declention rules
    # number we don't need look further.
    if f == AT_AF:
        return [f]

    len_f = len(f.split("_"))
    # If len_f is longer than three means the user has inputed a invalid string
    if len_f > 3:
        raise Exception(f"Sorry, the input string is ambigous")

    # Check wheter the user input e.g. the number twice
    input_value_types = []
    for v in f.split("_"):
        if v not in dec:
            raise Exception(
                f"Sorry, the input string is ambigous for the part {v} of the input string {f}"
            )

        if dec[v] in input_value_types:
            raise Exception(f"Duplicate value in the input string")
        else:
            input_value_types.append(dec[v])

    # if the length of len_f is three return the input in the correct order
    new_f = []
    if len_f == 3:
        new_f.append(f)

    # if the user only adds one or two values we need to find the category that is missing
    # and populate the string with those categories.
    categories_present = [dec[x] for x in f.split("_")]
    missing_categories = [x for x in correct_order if x not in categories_present]

    if len_f == 2:
        for x in [k for k, v in dec.items() if v == missing_categories[0]]:
            new_f.append(f"{f}_{x}")

    if len_f == 1:
        for m_one in [k for k, v in dec.items() if v == missing_categories[0]]:
            for m_two in [k for k, v in dec.items() if v == missing_categories[1]]:
                new_f.append(f"{f}_{m_one}_{m_two}")

    # Now lest sort all the enties in new_f
    for i in range(len(new_f)):
        new_f[i] = {dec[v]: v for v in new_f[i].split("_")}
        new_f[i] = "{}_{}_{}".format(
            new_f[i][correct_order[0]],
            new_f[i][correct_order[1]],
            new_f[i][correct_order[2]],
        )
    logging.debug(f"Sorted input string is {new_f[i]}")
    return new_f
